package admin;

import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;

import logging.Logger;
import reg.RegServer;
import scg.Constants;
import scg.Util;
import scg.game.Game;
import scg.game.GameI;
import scg.game.HistoryFile;
import scg.game.PlayerProxyI;
import scg.game.StatisticsCollector;
import scg.game.Tournament;
import scg.gen.Config;
import scg.gen.ParseException;
import scg.gen.PlayerSpec;
import scg.gen.PlayersFile;
import edu.neu.ccs.demeterf.http.server.Factory;
import edu.neu.ccs.demeterf.http.server.ServerThread;
import edu.neu.ccs.demeterf.lib.List;
import edu.neu.ccs.demeterf.lib.Option;
import edu.neu.ccs.demeterf.util.CLI;

/** TODO: 
 *    - (Maybe) Support overriding registrations, incase players die */

/**  */
public class AdminMain {

    /** Current Version */
    static final String REV = "$Rev: 537 $".substring(6, 9);

    
    /** Options */
    public static final String HELP = "--help", NO_REG = "--noreg", RELATIVE_TIME = "--relative",
            SAVE_PLAYERS = "--savereg", TOURNAMENT = "--tournament", PARTIAL_REG = "--partial",
            CAPTURE = "--capture", DONTWAIT = "--dontwait", EXCEPTIONS = "--exceptions",
            MULTIPLAYER = "--multiplayer", RESPONSELIMIT = "--responselimit", NOSTATUS = "--nostatus";

    static final List<String> allArgs = List.create(HELP, NO_REG, RELATIVE_TIME, SAVE_PLAYERS, TOURNAMENT, PARTIAL_REG,
            CAPTURE, DONTWAIT, EXCEPTIONS, MULTIPLAYER, RESPONSELIMIT, NOSTATUS);

    /** The list of global options */
    static List<String> globalOptions;
    /** See if the option has been set */
    static boolean optionSet(String s){
        return globalOptions.contains(s);
    }
    /** Get the Argument of an option like "--file:NAME" */
    static String optionArg(String opt){
        return firstOrEmpty(CLI.separateOption(opt, globalOptions));
    }
    
    /** Print usage information and Quit */
    static void usage(boolean full, String err){
        if (err.length() > 0) {
            p("  !! " + err + "\n");
        }
        p("  ** Usage: java AdminMain [Options] <Time> [PlayersFile]");
        if (full) {
            p("\n  Options can be:\n" 
                    + "      --help         Print this message\n"
                    + "      --relative     Use a 'relative' start time instead of an absolute\n"
                    + "                       start time. Used for running a quick competition.\n"
                    + "      --tournament   Run a tournament style competition\n"
                    + "\n    * Registration\n"
                    + "      --noreg        Skip registration, use a preset Players file\n"
                    + "      --savereg      Save the Registration into the Players file\n"
                    + "      --partial      Use the Players file as a partial registration\n"
                    + "\n    * Server\n"
                    + "      --dontwait           Quit as soon as the competition is done\n"
                    + "      --capture:<file>     Capture all the HTTP messages to the given file\n"
                    + "      --exceptions:<file>  Capture Exception contexts to the given file\n"
                    + "      --multiplayer:N      Allow each team up to 'N' registrations.  Registered\n"
                    + "                             Players will have their login number appended to\n"
                    + "                             their team name, e.g., 'MyTeam:2'\n"
                    + "      --responselimit:N    Limit Player responses to 'N' bytes. Default is unlimited.\n\n"
                    
                    + "    Absolute Time (default) is formated as: 'mm/dd/yyyy hh:mm:ss AM|PM EDT'.\n"
                    + "    Relative Time is an integer; the number of seconds before starting\n\n"
                    + "    The location of a (registered) players info file can be\n"
                    + "      given... otherwise the default is '" + Constants.PLAYERS_FILE + "'\n");
        } else {
            p("  ** Use --help for a full description\n");
        }
        System.exit(1);
    }

    /**  */
    public static void p(String s){
        System.err.println(s);
    }

    /** Run the Admin... */
    public static void main(String[] argArr) throws Exception{
        List<String> split[] = CLI.splitArgs(argArr), opts = split[CLI.OPTS], args = split[CLI.ARGS];
        globalOptions = opts;
        
        p("\n               /-------------------------------------\\");
        p("               |          The Administrator          |");
        p("               |             Version " + REV + "             |");
        p("               |            Game Rev " + Constants.REV + "             |");
        p("               \\-------------------------------------/\n");

        if (optionSet(HELP)) {
            usage(true, "");
        }
        List<String> unknown = CLI.invalidOptions(opts, allArgs);
        if (!unknown.isEmpty()) {
            usage(false, "Unknown Option(s): " + unknown.toString(", ", ""));
        }
        if (args.length() < 1) {
            usage(false, "Not enough manditory arguments");
        }

        String date = args.lookup(0);
        List<PlayerSpec> players;
        Date start;
        Logger logger = Logger.text(System.out, Util.logFileName("admin"));
        logger.notify("*");
        logger.notify("* Admin Version " + REV);
        logger.notify("*");
        try {
            if (optionSet(RELATIVE_TIME)) {
                Calendar cal = new GregorianCalendar();
                cal.setTime(Util.now());
                cal.add(Calendar.SECOND, Integer.parseInt(date));
                start = cal.getTime();
            } else {
                start = Util.parseDate(date);
            }

        } catch (java.text.ParseException pe) {
            usage(false, "Error parsing Date");
            return;
        }

        AdminStatus status = new AdminStatus(Constants.DEF_STAT_PORT, start, optionSet(NOSTATUS));
        ServerThread statusServer = optionSet(NOSTATUS)?null:Factory.create(status,false,20);

        try {
            // Capture the HTTP Req/Resp to a file
            String file = optionArg(CAPTURE);
            if (file.length() > 0) {
                RemotePlayerProxy.setSaveFile(file);
            }
            // Save HTTP/Parse Exceptions to the given File (includes Req/Resp)
            file = optionArg(EXCEPTIONS);
            if (file.length() > 0) {
                GameI.HTTPTransException.setSaveFile(file);
            }

            if (optionSet(NO_REG)) {
                logger.event("Skipping Registration");
                String pfile = args.length() > 1 ? args.lookup(1) : Constants.PLAYERS_FILE;
                players = loadPlayers(pfile);
            } else {
                String numLogins = optionArg(MULTIPLAYER);
                RegServer reg = RegServer.create(logger, optionSet(SAVE_PLAYERS) ? (args.length() > 1 ? args
                        .lookup(1) : Constants.PLAYERS_FILE) : "", optionSet(PARTIAL_REG),
                        numLogins.length() > 0 ? Integer.parseInt(numLogins) : 1);
                status.setReg(reg);
                players = reg.runRegistration(Option.some(start));
            }
            Util.waitUntil(start);
            status.finishReg();
        } catch (Exception e) {
            usage(false, "Exception while starting up: " + e.getClass().getName() + " : " + e.getMessage());
            e.printStackTrace();
            return;
        }
        Config config = loadConfig(Constants.CONFIG_FILE);
        int mc = config.getPredicate().getMaxClauses();
        String maxResp = optionArg(RESPONSELIMIT);
        List<PlayerProxyI> proxies = wrapPlayerSpecs(players, 
                maxResp.isEmpty()?0:Long.parseLong(maxResp)).shuffle();
        if (optionSet(TOURNAMENT)) {
            Tournament t = new Tournament(config, proxies);
            status.setTourn(t);
            HistoryFile history = new HistoryFile(Constants.HISTORY_FILE_PREFIX + "_tournament", start,
                    Constants.HISTORY_FILE_SUFFIX);
            logger.event("Tournament Competition Started");
            t.start(history);
            logger.event("Competition Complete");

        } else {
            Game game = new Game(config, proxies);
            status.setGame(game);
            HistoryFile history = new HistoryFile(Constants.HISTORY_FILE_PREFIX, start, Constants.HISTORY_FILE_SUFFIX);
            StatisticsCollector sc = new StatisticsCollector(Constants.HISTORY_FILE_PREFIX + "_statistics", start,
                    Constants.HISTORY_FILE_SUFFIX);
            logger.event("Group Competition Started");
            game.start(history, sc, true);
            logger.event("Competition Complete");
        }
        status.finishGame();
        status.writeResults(logger, optionSet(TOURNAMENT));
        if (!optionSet(DONTWAIT)) {
            System.err.print("\n ** Game Completed, Hit Enter To Close Server: ");
            System.in.read();
            if(statusServer != null)
                statusServer.shutdown();
        } else {
            if(statusServer != null)
                statusServer.shutdown(false);
        }
        System.exit(0);
    }

    static String firstOrEmpty(String[] ss){
        if (ss.length == 0) {
            return "";
        }
        return ss[0];
    }

    private static List<PlayerProxyI> wrapPlayerSpecs(List<PlayerSpec> playerSpecs, final long maxResp){
        return playerSpecs.map(new List.Map<PlayerSpec, PlayerProxyI>() {

            @Override
            public PlayerProxyI map(PlayerSpec spec){
                return new RemotePlayerProxy(spec, maxResp);
            }
        });
    }

    /**
     * Load playerSpecs from a file
     */
    public static List<PlayerSpec> loadPlayers(String playerFile) throws ParseException, java.io.IOException{
        return PlayersFile.parse(new java.io.FileInputStream(playerFile)).getPlayerSpecs();
    }

    /**
     * Load Config from a file
     */
    public static Config loadConfig(String configFile) throws ParseException, java.io.IOException{
        return Config.parse(new java.io.FileInputStream(configFile));
    }

}
